from dataclasses import dataclass, asdict
from datetime import datetime
from xxhash import xxh64

from typing import Union, Sequence, Mapping

from .errors import PinsVersionError
from ._types import StrOrFile, IOBase

VERSION_TIME_FORMAT = "%Y%m%dT%H%M%SZ"


class _VersionBase:
    pass


@dataclass
class VersionRaw(_VersionBase):
    version: str

    def to_dict(self) -> Mapping:
        return asdict(self)


@dataclass
class Version(_VersionBase):
    created: datetime
    hash: str

    def to_dict(self) -> Mapping:
        # properties not automatically added, so need to handle manually
        res = asdict(self)
        res["version"] = self.version

        return res

    @property
    def version(self) -> str:
        date_part = self.created.strftime(VERSION_TIME_FORMAT)
        hash_part = self.hash[:5]
        return f"{date_part}-{hash_part}"

    @staticmethod
    def parse_created(x):
        return datetime.strptime(x, VERSION_TIME_FORMAT)

    def render_created(self):
        return self.created.strftime(VERSION_TIME_FORMAT)

    @staticmethod
    def hash_file(f: IOBase, block_size: int = -1) -> str:
        # TODO: what kind of things implement the "buffer API"?
        hasher = xxh64()

        buf = f.read(block_size)

        while len(buf) > 0:
            hasher.update(buf)
            buf = f.read(block_size)

        return hasher.hexdigest()

    @classmethod
    def from_string(cls, version: str) -> "Version":
        parts = version.split("-")

        if len(parts) != 2:
            raise PinsVersionError(
                "version string can only have 1 '-', but contains %s" % len(parts)
            )

        dt_string, hash_ = parts

        # TODO: the datetime from pins is not timezone aware, but it looks like
        # R pins parses as UTC, then unsets the UTC part?
        try:
            created = cls.parse_created(dt_string)
        except ValueError:
            raise PinsVersionError("Invalid date part of version: " % dt_string)

        obj = cls(created, hash_)

        if obj.version != version:
            raise ValueError(
                "Version parsing failed. Received version string {version}, but "
                "output version is {cls.version}."
            )

        return obj

    @classmethod
    def from_files(
        cls, files: Sequence[StrOrFile], created: Union[datetime, None] = None
    ) -> "Version":
        hashes = []
        for f in files:
            hash_ = cls.hash_file(open(f, "rb") if isinstance(f, str) else f)
            hashes.append(hash_)

        if created is None:
            created = datetime.now()

        if len(hashes) > 1:
            raise NotImplementedError("Only 1 file may be currently be hashed")

        return cls(created, hashes[0])

    @classmethod
    def from_meta_fields(cls, created: str, hash: str):
        created_dt = cls.parse_created(created)
        return cls(created_dt, hash)


def guess_version(x: str):
    try:
        return Version.from_string(x)
    except PinsVersionError:
        return VersionRaw(x)



import sys

from functools import update_wrapper
from types import MethodType
from warnings import warn

from .config import pins_options


def inform(log, msg):
    if log is not None:
        log.info(msg)

    if not pins_options.quiet:
        print(msg, file=sys.stderr)


def warn_deprecated(msg):
    warn(msg, DeprecationWarning)


class ExtendMethodDoc:
    # Note that the indentation assumes these are top-level method docstrings,
    # so are indented 8 spaces (after the initial sentence).
    template = """\
{current_doc}

        Parent method documentation:

        {parent_doc}
        """

    def __init__(self, func):
        self.func = func

        # allows sphinx to add the method signature to the docs
        # this is pretty benign, since it's very hard to call a descriptor
        # after class initialization (where __set_name__ is called).
        self.__call__ = func

    def __set_name__(self, owner, name):
        bound_parent_meth = getattr(super(owner, owner), name)

        self._parent_doc = bound_parent_meth.__doc__
        self._orig_doc = self.func.__doc__

        if self._orig_doc is not None:
            # update the docstring of the subclass method to include parent doc.
            self.func.__doc__ = self.template.format(
                current_doc=self._orig_doc, parent_doc=self._parent_doc
            )

        # make descriptor look like wrapped function
        update_wrapper(
            self, self.func, ("__doc__", "__name__", "__module__", "__qualname__")
        )

    def __get__(self, obj, objtype=None):
        if obj is None:
            # accessing from class, return descriptor itself.
            return self

        # accessing from instance
        return MethodType(self.func, obj)

    def __call__(self, *args, **kwargs):
        # this is defined, so that callable(ExtendMethodDoc(...)) is True,
        # which allows all the inspect machinery to give sphinx the __call__
        # attribute we set in __init__.
        raise NotImplementedError()



from pins.versions import Version
from datetime import datetime

import xxhash
import pytest

from io import BytesIO

EXAMPLE_DATE = datetime(2021, 1, 2, 13, 58, 59)


@pytest.fixture
def bytes_():
    return BytesIO(b"123"), xxhash.xxh64(b"123").hexdigest()


def test_version_from_string():
    version = Version.from_string("20220209T220116Z-baf3f")
    assert str(version.created) == "2022-02-09 22:01:16"
    assert version.hash == "baf3f"


def test_version_hash_file(bytes_):
    f_bytes, digest = bytes_
    assert Version.hash_file(f_bytes) == digest


def test_version_from_files(bytes_):
    f_bytes, digest = bytes_
    v = Version.from_files([f_bytes], EXAMPLE_DATE)

    assert v.hash == digest
    assert v.created == EXAMPLE_DATE



# flake8: noqa

# Set version ----
from importlib_metadata import version as _v

__version__ = _v("pins")

del _v


# Imports ----
from .cache import cache_prune, cache_info
from .constructors import (
    board_deparse,
    board_folder,
    board_temp,
    board_local,
    board_github,
    board_urls,  # DEPRECATED
    board_url,
    board_rsconnect,
    board_azure,
    board_s3,
    board_gcs,
    board,
)



import pandas as pd

from importlib_resources import files

from pins import board_folder

OLD_BOARD = files("pins") / "tests" / "pins-old-types"
DST_DF = pd.DataFrame({"a": [1, 2], "b": ["x", "y"]})


def test_compat_old_types_load_table():
    board = board_folder(OLD_BOARD)
    src_df = board.pin_read("a-table")

    assert isinstance(src_df, pd.DataFrame)
    assert src_df.equals(DST_DF)



import pytest
import datetime


from pins.errors import PinsError
from pins.tests.helpers import xfail_fs
from pins.tests.conftest import PATH_TO_EXAMPLE_BOARD, PATH_TO_MANIFEST_BOARD


NOT_A_PIN = "not_a_pin_abcdefg"
PIN_CSV = "df_csv"

# set up board ----


@pytest.fixture(scope="session")
def board(backend):
    board = backend.create_tmp_board(str(PATH_TO_EXAMPLE_BOARD.absolute()))

    yield board

    backend.teardown_board(board)


@pytest.fixture(scope="session")
def board_manifest(backend):
    # skip on rsconnect, since it can't add a manifest and the pin names
    # are too short for use to upload (rsc requires names > 3 characters)
    if backend.fs_name == "rsc":
        pytest.skip()

    board = backend.create_tmp_board(str(PATH_TO_MANIFEST_BOARD.absolute()))

    yield board

    backend.teardown_board(board)


# pin_list --------------------------------------------------------------------


def test_compat_pin_list(board):
    src_sorted = sorted(board.pin_list())
    dst_sorted = ["df_arrow", "df_csv", "df_rds", "df_unversioned"]

    if board.fs.protocol == "rsc":
        # rsc backend uses <user_name>/<content_name> for full name
        dst_sorted = [f"{board.user_name}/{content}" for content in dst_sorted]

    assert src_sorted == dst_sorted


# pin_versions ----------------------------------------------------------------


def test_compat_pin_versions(board):
    if board.fs.protocol == "rsc":
        pytest.skip("RSC uses bundle ids as pin versions")
    versions = board.pin_versions("df_csv", as_df=False)
    v_strings = list(v.version for v in versions)
    assert v_strings == ["20220214T163718Z-eceac", "20220214T163720Z-9bfad"]


@pytest.mark.skip("Used to diagnose os listdir ordering")
def test_compat_os_listdir():
    import os

    res = os.listdir(PATH_TO_EXAMPLE_BOARD / "df_csv")
    dst = ["20220214T163718Z-eceac", "20220214T163720Z-9bfad"]

    assert res == dst


# pin_exists --------------------------------------------------------------------


def test_compat_pin_exists_succeed(board):
    assert board.pin_exists(PIN_CSV)


def test_compat_pin_exists_fails(board):
    assert board.pin_exists(NOT_A_PIN) is False


# pin_meta --------------------------------------------------------------------


def test_compat_pin_meta(board):
    # Note that this fetches the latest of 2 versions
    meta = board.pin_meta(PIN_CSV)

    if board.fs.protocol == "rsc":
        # TODO: afaik the bundle id is largely non-deterministic, so not possible
        # to test, but should think a bit more about it.
        assert meta.name == "derek/df_csv"
    else:
        assert meta.version.version == "20220214T163720Z-9bfad"
        assert meta.version.created == datetime.datetime(2022, 2, 14, 16, 37, 20)
        assert meta.version.hash == "9bfad"

        assert meta.name == "df_csv"

    assert meta.title == "df_csv: a pinned 2 x 3 data frame"
    assert meta.description is None
    assert meta.created == "20220214T163720Z"
    assert meta.file == "df_csv.csv"
    assert meta.file_size == 28
    assert meta.pin_hash == "9bfad6d1a322a904"
    assert meta.type == "csv"

    # TODO(question): coding api_version as a yaml float intentional?
    assert meta.api_version == 1.0
    assert meta.user == {}


def test_compat_pin_meta_pin_missing(board):
    with pytest.raises(PinsError) as exc_info:
        board.pin_meta(NOT_A_PIN)

    assert f"{NOT_A_PIN} does not exist" in exc_info.value.args[0]


@xfail_fs("rsc")
def test_compat_pin_meta_version_arg(board):
    # note that in RSConnect the version is the bundle id
    meta = board.pin_meta(PIN_CSV, "20220214T163718Z-eceac")
    assert meta.version.version == "20220214T163718Z-eceac"
    assert meta.version.hash == "eceac"


def test_compat_pin_meta_version_arg_error(board):
    bad_version = "123"
    with pytest.raises(PinsError) as exc_info:
        board.pin_meta(PIN_CSV, bad_version)

    msg = exc_info.value.args[0]
    assert PIN_CSV in msg
    assert bad_version in msg


# pin_read ----


def test_compat_pin_read(board):
    import pandas as pd

    p_data = PATH_TO_EXAMPLE_BOARD / "df_csv" / "20220214T163720Z-9bfad" / "df_csv.csv"

    src_df = board.pin_read("df_csv")
    dst_df = pd.read_csv(p_data)

    assert isinstance(src_df, pd.DataFrame)
    assert src_df.equals(dst_df)


def test_compat_pin_read_supported(board):
    with pytest.raises(NotImplementedError):
        board.pin_read("df_rds")


# pin_write ----

# manifest -----


def test_board_pin_write_manifest_name_error(board_manifest):
    if board_manifest.fs.protocol == "rsc":
        pytest.skip()

    with pytest.raises(ValueError) as exc_info:
        board_manifest.pin_write([1], "_pins.yaml", type="json")

    assert "name '_pins.yaml' is reserved for internal use." in exc_info.value.args[0]


def test_board_manifest_pin_list_no_internal_name(board_manifest):
    assert set(board_manifest.pin_list()) == {"x", "y"}


def test_board_manifest_pin_exist_internal_name_errors(board_manifest):
    with pytest.raises(ValueError) as exc_info:
        board_manifest.pin_exists("_pins.yaml")

    assert "reserved for internal use." in exc_info.value.args[0]


def test_board_manifest_pin_read_internal_errors(board_manifest):
    with pytest.raises(ValueError) as exc_info:
        board_manifest.pin_read("_pins.yaml")

    assert "reserved for internal use." in exc_info.value.args[0]


def test_board_manifest_pin_search(board_manifest):
    res = board_manifest.pin_search("x", as_df=False)

    assert len(res) == 1
    assert res[0].name == "x"
